/*
 *  Copyright (c) 2014 The WebRTC project authors. All Rights Reserved.
 *
 *  Use of this source code is governed by a BSD-style license
 *  that can be found in the LICENSE file in the root of the source
 *  tree. An additional intellectual property rights grant can be found
 *  in the file PATENTS.  All contributing project authors may
 *  be found in the AUTHORS file in the root of the source tree.
 */

namespace webrtc {
namespace videocapturemodule {

// Ensure any necessary initialization of webrtc::videocapturemodule has
// completed.
void EnsureInitialized();

}  // namespace videocapturemodule.
}  // namespace webrtc.
