/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

/* avoids contributing to the min width when Lightning is not selected */
#calendarTabPanel:not([selected]) {
  visibility: collapse;
}

.imipMoreButton >.toolbarbutton-icon {
  display: none;
}
