/* Any copyright is dedicated to the Public Domain.
 * http://creativecommons.org/publicdomain/zero/1.0/ */

#ifndef GRAPHIC_BUFFER_H_
#define GRAPHIC_BUFFER_H_

namespace stagefright {

class GraphicBuffer : public RefBase {
};

}

#endif
