/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

MFPLAT_FUNC(MFCreateSample, "mfplat.dll");
MFPLAT_FUNC(MFCreateAlignedMemoryBuffer, "mfplat.dll");
MFPLAT_FUNC(MFGetStrideForBitmapInfoHeader, "evr.dll");
MFPLAT_FUNC(MFCreateMediaType, "mfplat.dll");