/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

:root {
  --toolbarHighlight: rgba(253, 253, 253, 0.45);
  --fgTabTexture: linear-gradient(transparent, transparent 2px, rgba(254, 254, 254, 0.72) 2px, rgba(254, 254, 254, 0.72) 2px, rgba(250, 250, 250, 0.88) 3px, rgba(250, 250, 250, 0.88) 3px, rgba(254, 254,  254, 0.72) 4px, rgba(254, 254, 254, 0.72) 4px, var(--toolbarHighlight));
  --fgTabBackgroundColor: -moz-dialog;
  --fgTabTextureLWT: var(--fgTabTexture);

  /* Aero-only variables */

  --customToolbarColor: hsl(210, 75%, 92%);
}
