/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

statusbar {
  font-weight: bold;
  height: 2em;
}
#security-button {
  width: 20px;
  padding: 0;
  margin: 0;
}
#security-button[level="high"],
#security-button[level="low"] {
  background-image: url("chrome://chat/skin/icons/secure.png");
}

#security-button[level="broken"] {
  background-image: url("chrome://chat/skin/icons/insecure.png");
}

#header {
  padding: 5px;
  border-bottom: 1px solid black;
}

#header label {
  font-weight: bold;
  font-size: 1.4em;
}

#header description {
  padding: 6px;
  font-weight: bold;
}
